#ifndef __KERNSH_H
#define __KERNSH_H

#include <linux/module.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <sys/utsname.h>
#include <fcntl.h>
#include <elf.h>
#include <linux/types.h>
#include <asm/segment.h>
#include <unistd.h>
#include <linux/sys.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <string.h>
#include <time.h>
#include <ctype.h>
#include <stdarg.h>
#include <sched.h>
#include <linux/kernel.h>
#include <readline/readline.h>
#include <readline/history.h>
#include <sys/mman.h>
#include <asm/unistd.h>

#define PROMPT "\033[32m%\033[0m "    
#define KMEM "/dev/kmem"

#ifndef NR_syscalls
#define NR_syscalls 256
#endif

#ifndef EM_486
#define EM_486 6
#endif

#define CHECK_NUMBER_SYSCALL( n ) { \
		if ( n < 0 || n > NR_syscalls-1 ) { \
			fatal( "syscall # (%i) not in range\n", n ); \
			return; \
		} \
}


/* global variables used by all source files */
/* file descriptor of /dev/kmem */
extern int file_descriptor_kmem;
/* sys_call_table address */ 
extern size_t p_sys_call_table; 
/* system_call address */ 
extern size_t p_system_call; 
/* table of syscalls */
extern unsigned long sys_call_table[ NR_syscalls ];
/* modules supported or not */
extern int modules_not_supported;


#define XSTRTOUL( string, addr ) { \
			addr = strtoul( string, NULL, 16 ); \
			if ( addr < 0 ) { \
				fatal( "please, a valid address..\n\n" ); \
				return; \
			} \
		}
	
#define NOARGUMENTS( n, pargc, name ) { \
			if ( pargc == n ) { \
				fatal( "no arguments presents..\n\n" ); \
				fatal( "Type \"help\" followed by command name for full documentation.\n\n" ); \
				return; \
			} \
		}

/* prototypes */
int query_module( const char * name, int which, void * buf, 
size_t bufsize, size_t * ret );
int get_kernel_syms( struct kernel_sym * table );
void do_module( int argc, char * argv[] );
void do_exit( int argc, char * argv[] );
void do_symbol( int argc, char * argv[] );
void do_syscall( int argc, char * argv[] );
void do_kmem( int argc, char * argv[] );
void do_idt( int argc, char * argv[] );
void do_display_system_info( int argc, char * argv[] );
void do_bzimage( int argc, char * argv[] );

/* usage */
void usage_symbol( void );
void usage_syscall( void );
void usage_module( void );
void usage_kmem( void );
void usage_idt( void );
void usage_display_system_info( void );
void usage_bzimage( void );

typedef struct command_s {
	char * name;
	void (* func)( int argc, char * argv[] );
	char * description;
	void (* usage)( void );
} command_t;

/* to add a new command, enter it in this table */
static const command_t p_command_main[] =
{
	{ "display", do_display_system_info, "System information", 
	usage_display_system_info },
	{ "symbol", do_symbol, "Description of the symbols", usage_symbol },
	{ "syscall", do_syscall, "Description of the syscall", usage_syscall },
	{ "module", do_module, "Management of the modules", usage_module },
	{ "kmem", do_kmem, "Use of /dev/kmem", usage_kmem },
	{ "idt", do_idt, "Handle Interrupt Description Table", usage_idt },
	{ "bzimage", do_bzimage, "Handle compressed kernel image", usage_bzimage },
	{ "help", 0, "Display this help", 0 },
	{ "exit", do_exit, "Quit kernsh", 0 },
	{ NULL, 0, NULL, 0 }
};

/* kmem.c */
int open_kmem( void );
void close_kmem( int kd );
int write_kmem( void * buf, size_t offset, int size, int kd );
int read_kmem( void * buf, size_t offset, int size, int kd );
void kdump( size_t addr, int ssize, int fd );

/* symbol.c */
size_t get_symbol_by_kernel_sym( char * symbol );
size_t get_symbol_by_query_module( char * symbol );
size_t get_symbol_by_system_map( char * symbol, int regexp );
size_t get_symbol_by_name( char * symbol );
char * get_symbol_by_addr( char * addr );
size_t get_next_symbol_by_addr( size_t addr );
int found_vmlinux( void );

/* misc.c */
char * dupstr( char * s );
void fatal( const char * format, ... );
void hexdump( unsigned char * data, unsigned int amount );
int parse( char * s1, char * s2, size_t addr );
char enter( void );
char * stripwhite( char * string );
char * string2hex( char * str, int * nu );
void * xmalloc( size_t size );

/* idt.c */
struct s_idtr {
    unsigned short limit;
    unsigned int base;
} __attribute__ ((packed))idtr;

struct {
    unsigned short off1;
    unsigned short sel;
    unsigned char none;
	unsigned char flags;
    unsigned short off2;
} __attribute__ ((packed))idt;

unsigned char * get_idt_segment( unsigned int interrupt );

/* syscall.c */
void call_syscall( int number );

/* md5.c */
struct MD5Context {
	unsigned long buf[ 4 ];
	unsigned long bits[ 2 ];
	unsigned char in[ 64 ];
};

void MD5Init( struct MD5Context * context );
void MD5Update( struct MD5Context * context, unsigned char const * buf,
unsigned len );
void MD5Final( unsigned char digest[ 16 ], 
struct MD5Context * context );
void MD5Transform( unsigned long buf[ 4 ], 
const unsigned char in[ 64 ] );
void dumpmd5( char * pass, int value, 
unsigned char md5buffer[ BUFSIZ ] );

/* signature.c */
#define SIG_FOR_SYSCALL 0
#define SIG_FOR_IDT 1

void create_file_signatures( char * banner, char * signatures_file, 
int size_sign, int family );
int check_file_signatures( char * banner, char * signatures_files, 
int size_sign, int family );

/* disasm.c */
int disasm( size_t start_address, int len );

/* module.c */
typedef struct _module {
	Elf32_Ehdr ehdr;
	Elf32_Shdr * shdr;
	unsigned long maddr;
	int	len;
	int	strtabidx;
	char ** section;
} module_t;

int load_module( module_t * module, int fd );
unsigned long init_module( module_t * p_module, unsigned long maddr );
unsigned long find_symaddr( module_t * p_module, const char * symbol );
unsigned long make_module( module_t * module, unsigned long dst );
int module_size( int fd );
int kinsert( char * m, size_t location, char ** symbol, 
char * function, size_t * func_addr );
void * rel_module(char * filename, unsigned long addr, int * len, 
char ** sym, char * function, size_t * func_addr);

#endif /* __KERNSH_H */

